import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * This example demonstrates the "producer/consumer" pattern for
 * parallel programs. The producer creates work for the consumer
 * to process and puts it into a queue; the consumer loops
 * infinitely, waiting for something to appear in the queue and
 * then processes it.
 */
public class E_ProducerConsumer {
  public static void main(String[] args) {
    Queue<Long> queue = new ConcurrentLinkedQueue<Long>();
    Thread producerThread = new Thread(new PrimeProducer(queue, 189877));
    Thread consumerThread = new Thread(new NumberPrinter(queue, "is prime!"));

    producerThread.start();
    consumerThread.start();
  }
}

/**
 * Produce prime numbers and put them in a processing
 * queue.
 */
class PrimeProducer implements Runnable {
  Queue<Long> queue;
  long num;

  /**
   * Create a producer that puts items in the given queue.
   * Prime numbers will start at `start`.
   */
  public PrimeProducer(Queue<Long> queue, long start) {
    this.queue = queue;
    this.num = start;
  }

  /**
   * Check whether a single number is prime.
   */
  private boolean isPrime(long num) {
    for (long i = 2; i < num; ++i) {
      if (num % i == 0) {
        return false;
      }
    }
    return true;
  }

  public void run() {
    while (true) {
      if (isPrime(num)) {
        // Send the prime to the consumer.
        queue.add(num);
      }

      // Try the next number.
      num++;
    }
  }
}

/**
 * A consumer that reads numbers from a queue and prints them.
 */
class NumberPrinter implements Runnable {
  Queue<Long> queue;
  String message;

  /**
   * Create a printer that reads numbers from queue and prints
   * them followed by message.
   */
  public NumberPrinter(Queue<Long> queue, String message) {
    this.queue = queue;
    this.message = message;
  }

  public void run() {
    while (true) {
      // Get a prime from the producer.
      synchronized (queue) {
        if (!queue.isEmpty()) {
          long p = queue.remove();
          System.out.println(p + " " + message);
        }
      }
    }
  }
}
